"""Preprocessing of images
==========================

Images acquired from experiments can suffer from noise, reflection and other
undesirable effects. Preprocessing images by careful use of thresholds and
filters can significantly improve the quality of PIV.

Provides:

.. autosummary::
   :toctree:

   base
   toolbox
   image2image

"""

from .base import PreprocBase

__all__ = ["PreprocBase"]
