"""Reconstruction algorithms
============================

In general reconstruction simply means applying the calibration of mutiple
cameras to obtain a sense of depth. It is performed prior to PIV in the case of
tomographic PIV (3D-3C PIV) and after PIV in the case of stereo PIV (2D-3C
PIV).

Provides:

.. autosummary::
   :toctree:

   tomo

"""
