"""Graphical user interface
===========================

Two simple graphical user interfaces (GUI) can be launched with the commands
`fluidimviewer` and `fluidimlauncher`.

This sub-package contains the code for the fluidimage GUI:

.. autosummary::
   :toctree:

   imviewer
   launcher
   pg_main

"""
