"""Utilities for numerical computation
======================================

This subpackage contains specialized functions and classes for efficient
calculations.

.. autosummary::
   :toctree:

   fft
   correl
   subpix
   interpolate
   smooth_clean

"""
